﻿////
/// Copyright(C) 2020 Micronet Co. All Rights Reserved.
//
using MQTTnet;
using MQTTnet.Client;
using MQTTnet.Client.Options;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Threading.Tasks;

namespace Rsi
{
    public partial class HiMQTT : Rsi
    {
        //--------------------------------------------------
        // クラスメンバ変数の定義
        //--------------------------------------------------
        private InfomationArea Inf_Table;       // 情報領域テーブル

        /// MQTT protocol
        private MqttFactory factory;            // MQTT クライアントインスタンス
        private IMqttClient mqttClient;
        // 通信監理フレームワーク
        private Communication comm;

        // Properties
        private string ServerAddr;           // MQTT ブローカーアドレス
        private List<string> SubscTopics;    // MQTT サブスクリプション　トピック名リスト
        // Properties（オプション）
        private int QoSLevel;                // QoSレベル（デフォルト0)
        private bool UseWebsock;             // true=
        private bool UseCredentials;         // true=
        private string User;
        private string Pass;
        private bool UseClientId;            // true=
        private string ClientId;             // Client ID
        // Properties（現バージョンでは非サポート）
        private bool UseTls;                 // true=
        private bool UseQos;                 // true=
        private bool UseRetain;              // true=

        // Statistics
        public int ConnectCounter = 0;      // 接続成功回数
        public int DisconnectCounter = 0;   // 切断発生回数

        // const values
        public const int MQTT_PORT = 1883;          // MQTT protocol default port number
        public const int MQTT_CONN_TCP = 1;         // Broker TCP
        public const int MQTT_CONN_WEBSOCK = 2;     // Broker Websocket
        public const int MQTT_CONN_TCP_TLS = 3;     // Broker TCP with TLS
        public const int MQTT_CONN_WEBSOCK_TLS = 4; // Broker Websocket with TLS
        public const int MQTT_INTERVAL = 300;
        public const int MQTT_PRIORITY = 0;

        //--------------------------------------------------
        // ステータスコード宣言
        //--------------------------------------------------

        /* 区分:その他のエラー */
        public const UInt32 ER_HI_CONF_USERNAME = 0x80000005;       // ユーザー名未指定エラー
        public const UInt32 ER_HI_CONF_PASSWORD = 0x80000006;       // パスワード未指定エラー
        public const UInt32 ER_HI_CONF = 0x80000007;                // 不明な設定項目が指定された
        public const UInt32 ER_HI_SIZE_OVER = 0x8000000D;           // 指定サイズオーバーエラー
        public const UInt32 ER_HI_BUFFER_NULL = 0x8000000E;         // バッファNULL指定エラー
        public const UInt32 ER_HI_BUFFERSIZE_SMALL = 0x8000000F;    // バッファサイズが読み込みサイズよりも小さいエラー
        public const UInt32 ER_HI_CONF_NULL = 0x80000010;           // configにnullが指定された
        public const UInt32 ER_HI_TARGET_NULL = 0x80000011;         // targetにnullが指定された
        public const UInt32 ER_HI_SIZE_ZERO = 0x80000012;           // 指定サイズ0エラー


        /// <summary>
        /// コンストラクタ
        /// </summary>
        public HiMQTT() : base()
        {
            Inf_Table = new InfomationArea("uri", "user", "pass");                 // 情報領域テーブルの生成・初期化

            // MQTTnet コンポーネントのインスタンスを生成します
            factory = new MqttFactory();
            mqttClient = factory.CreateMqttClient();

            // デフォルト設定値
            ServerAddr = "localhost:" + MQTT_PORT.ToString();
            SubscTopics = new List<string>();
            QoSLevel = 0;
            UseWebsock = false;
            UseCredentials = false;
            UseClientId = false;
            UseTls = false;
            UseQos = false;
            UseRetain = false;

            // 通信監理フレームワークの初期化
            comm = new Communication();
        }

        /// <summary>
        /// MQTTイベント処理を登録する
        /// </summary>
        /// <returns></returns>
        public bool fnMqttEvent()
        {
            ///
            /// 切断イベント到達処理
            ///
            mqttClient.UseDisconnectedHandler(e =>
            {
                Trace.WriteLine("### DISCONNECTED FROM SERVER ###");
                // 切断発生回数を増やす
                if (ConnectCounter > 0)
                    DisconnectCounter++;
                // 通信状態の制御
                comm.DisconnectedEvent();
            });

            ///
            /// メッセージ到達イベント到達処理
            ///
            mqttClient.UseApplicationMessageReceivedHandler(e =>
            {
                Trace.WriteLine("### RECEIVED APPLICATION MESSAGE ###");
                // MQTTメッセージの取り出し
                var MqttMessage = e.ApplicationMessage;
                // MQTTトピックを取り出す
                var topicname = MqttMessage.Topic;
                // 受信したペイロード(byte[])をstringに変換します
                var payload = Encoding.UTF8.GetString(MqttMessage.Payload);
                // 受信キューに書出し
                string item;
                if (comm.Rx.Count > 0)
                {
                    // 組み立て（未読がある場合は区切り文字'|'を付与してから）
                    item = "|" + topicname + "=" + payload;
                }
                else
                {
                    // 組み立て
                    item = topicname + "=" + payload;
                }
                // 受信キューに書出し
                byte[] Rx = Encoding.GetEncoding("shift_jis").GetBytes(item);
                comm.RxPush(Rx);
            });

            ///
            /// 接続成功イベント到達処理
            ///
            mqttClient.UseConnectedHandler(async e =>
            {
                Trace.WriteLine("### CONNECTED WITH SERVER ###");
                // サブスクリプション対象リストを使ってブローカに申請
                foreach (var i in SubscTopics)
                {
                    // MQTT Topicの組み立て
                    var topic = new MqttTopicFilterBuilder().WithTopic(i).Build();
                    // MQTTトピックのサブスクリプションをbrokerに通達
                    await mqttClient.SubscribeAsync(topic);
                    Trace.WriteLine(string.Format("### SUBSCRIBED {0} ###", i));
                }
                // 通信状態の制御
                ConnectCounter++;
                comm.ConnectedEvent();
            });

            return true;
        }

        /// <summary>
        /// ターゲットパラメータの解釈処理
        /// </summary>
        /// <param name="target">ターゲットパラメータ文字列</param>
        /// <returns></returns>
        internal UInt32 fnTreatTarget(string target)
        {
            UInt16 count = 0;
            foreach (string s in ItemValues.StringSplitOptions(target, '|')) // '|' で区切られたtarget文字列を処理する
            {
                // 一番先頭の文節は接続先として認識されます
                if (count++ == 0)
                {
                    ServerAddr = s;
                    Inf_Table.uri = ServerAddr;
                    continue;
                }

                // その他の文節はサブスクリプショントピック名として処理していきます。"item=value"の書式となる書式は将来拡張用で実際には"item"しか使いません。
                ItemValue setting = new ItemValue(s);
                SubscTopics.Add(setting.item);
                continue;
            }
            return ER_HI_OK;
        }

        /// <summary>
        /// コンフィグパラメータの解釈処理
        /// </summary>
        /// <param name="config">コンフィグパラメータ文字列</param>
        /// <returns></returns>
        internal UInt32 fnTreatConfig(string config)
        {
            foreach (string s in ItemValues.StringSplitOptions(config, '|')) // '|' で区切られたconfig文字列を処理する
            {
                // "item=value"の書式となるオプション設定値があれば処理していきます
                ItemValue setting = new ItemValue(s);
                if (setting.item.Equals("UseWebSocket"))
                {
                    UseWebsock = true;
                    continue;
                }
                if (setting.item.Equals("ClientId"))
                {
                    UseClientId = true;
                    ClientId = setting.value;
                    continue;
                }
                if (setting.item.Equals("User"))
                {
                    UseCredentials = true;
                    User = setting.value;
                    Inf_Table.username = setting.value;
                    continue;
                }
                if (setting.item.Equals("Pass"))
                {
                    UseCredentials = true;
                    Pass = setting.value;
                    Inf_Table.password = setting.value;
                    continue;
                }
            }
            return ER_HI_OK;
        }

        /// <summary>
        /// 初期化の要求
        /// </summary>
        /// <returns></returns>
        internal bool fnInitRequest()
        {
            fnMqttEvent();
            return true;
        }

        /// <summary>
        /// 初期化の完了待ち
        /// </summary>
        /// <param name="dwTimeout">msec</param>
        /// <returns></returns>
        internal int fnInitWait(UInt32 dwTimeout)
        {
            return 0;
        }

        /// <summary>
        /// 接続のリクエスト
        /// </summary>
        /// <returns></returns>
        internal bool fnOpenRequest()
        {
            if (DisconnectCounter > 0)
            {
                // 現在のステータス
                comm.ConnectingEvent();

                // 再接続
                Task.Run(() => mqttClient.ReconnectAsync());
                return true;
            }

            var options = new MqttClientOptionsBuilder();

            if (UseWebsock)
            {
                // WebSocket接続の使用
                options.WithWebSocketServer(ServerAddr);
            }
            else
            {
                // TCP接続の使用、Brokerアドレス の解釈
                var Remote = comm.TreatIPAddress(ServerAddr);
                options.WithTcpServer(Remote.Address.ToString(), Remote.Port);
            }

            try
            {
                // クライアントIdの指定あり
                if (UseClientId)
                    options.WithClientId(ClientId);

                // 接続ユーザ情報の指定あり
                if (UseCredentials)
                    options.WithCredentials(User, Pass);

                // セキュアTLSの指定あり
                if (UseTls)
                    options.WithTls();

                // クリーンなセッションを張ります
                options.WithCleanSession();

                // 現在のステータス
                comm.ConnectingEvent();

                // Brokerへの接続を要求
                mqttClient.ConnectAsync(options.Build(), CancellationToken.None).Wait();

                return true;
            }
            catch
            {
                return false;
            }
        }
        
        /// <summary>
        /// 接続完了するまで待ち
        /// </summary>
        /// <param name="dwTimeout">msec、WAIT_FOREVERは無限待ち</param>
        /// <returns></returns>
        internal int fnOpenWait(UInt32 dwTimeout)
        {
            int inttimeout = (dwTimeout > (UInt32)Int32.MaxValue) ? -1 : Convert.ToInt32(dwTimeout);
            return comm.ConnectedWait(inttimeout);
        }

        /// <summary>
        /// 切断のリクエスト
        /// </summary>
        /// <returns></returns>
        internal bool fnCloseRequest()
        {
            // 現在のステータス
            comm.DisconnectedEvent();
            // 切断
            Task.Run(() => mqttClient.DisconnectAsync());
            return true;
        }

        /// <summary>
        /// 切断完了の待ち
        /// </summary>
        /// <param name="dwTimeout">msec</param>
        /// <returns></returns>
        internal int fnCloseWait(UInt32 dwTimeout)
        {
            return 0;
        }

        /// <summary>
        /// MQTTブローカーから受信したトピック名と値のリストを受信キューから取り出します
        /// </summary>
        /// <param name="addr"></param>
        /// <param name="ipublist"></param>
        /// <param name="isize"></param>
        /// <returns></returns>
        internal int fnReadRequest( ref byte[] ipublist, int isize)
        {
            // 切断されている場合
            if (comm.Connected == false)
                return -2;
                
            // 受信キューから取り出します
            byte[] input = comm.RxPop(0 , ipublist.Length- 1);
            input.CopyTo(ipublist, 0);

            // 終端文字を付与します
            ipublist[input.Length] = 0x00;
            return input.Length;
        }

        /// <summary>
        /// 受信又は指定サイズの受信、あるいは切断されるまで待ち
        /// </summary>
        /// <param name="dwTimeout">msec</param>
        /// <returns></returns>
        internal int fnReadWait(UInt32 dwTimeout, int size)
        {
            int inttimeout = (dwTimeout > (UInt32)Int32.MaxValue) ? -1 : Convert.ToInt32(dwTimeout);
            return comm.RxWait(inttimeout, (size == 0) ? 1 : size);
        }

        /// <summary>
        /// トピック名と値のリストを元にMQTTブローカーにパブリッシュします
        /// </summary>
        /// <param name="addr">仮想I/O空間アドレス(0x00002000～0x00003FFF)</param>
        /// <param name="ipublist">'topic=value"、区切り文字'|'の書式で複数のPublishデータを文字列で渡す(sjis)</param>
        /// <param name="isize">Publishデータ参照バイトサイズ0～0x2000</param>
        /// <returns></returns>
        internal int fnWriteRequest( byte[] ipublist, int isize)
        {
            // 文字列に変換
            string pubs = System.Text.Encoding.GetEncoding("shift_jis").GetString(ipublist);

            // MQTTメッセージの組み立て
            var messagebuilder = new MqttApplicationMessageBuilder();

            // 出力予約リスト(buf)にたまっているだけ回す
            foreach (string s in ItemValues.StringSplitOptions(pubs, '|'))  // '|' で区切られた文字列を処理する
            {
                // TOPIC名と値の分解 "item=value"の書式
                ItemValue topic = new ItemValue(s);

                // TOPIC名と値の積み込み
                messagebuilder.WithTopic(topic.item);
                messagebuilder.WithPayload(topic.value);

                // QoSの指定がある場合
                if (UseQos)
                    messagebuilder.WithQualityOfServiceLevel(QoSLevel);

                // Retainの指定がある場合
                if (UseRetain)
                    messagebuilder.WithRetainFlag(true);

                // MQTTメッセージのビルド
                var messagelist = messagebuilder.Build();

                // MQTTメッセージの送出
                Task.Run(() => mqttClient.PublishAsync(messagelist));
            }

            return 0;
        }

        /// <summary>
        /// 書き込み完了の待ち
        /// </summary>
        /// <param name="dwTimeout">msec</param>
        /// <returns></returns>
        internal int fnWriteWait(UInt32 dwTimeout)
        {
            // MQTTでは書き込み保証がありませんので待ちません
            return 0;
        }
        /// <summary>
        /// 256バイト固定のbyte配列を返す
        /// </summary>
        /// <param name="src"></param>
        /// <returns></returns>
        byte[] GetFixedByteArray(byte[] src)
        {
            // 256バイト固定の配列を作成
            byte[] fixedSizeArray = new byte[256];
            // byteArrayの内容をfixedSizeArrayにコピー
            Array.Copy(src, fixedSizeArray, Math.Min(src.Length, fixedSizeArray.Length));
            return fixedSizeArray;
        }
        internal int fnReadInfo( ref byte[] buffer ,int offs ,int size )
        {
            try
            {
                //--------------------------------------------------
                // 管理データの組み立て
                //--------------------------------------------------
                List<byte> Info = new List<byte>();
                // 接続ステータス(0x4000 ,4)
                var status = (comm.Connected) ? 2 : (comm.Connecting) ? 1 : 0;
                Info.AddRange(BitConverter.GetBytes(status));

                // 接続先Uri(Sjis) 256byte (0x4004 , 256)
                var encoding = Encoding.GetEncoding("Shift-JIS");
                Info.AddRange(GetFixedByteArray(encoding.GetBytes(Inf_Table.uri)));

                // ユーザ名 256byte (0x4104 , 256)
                Info.AddRange(GetFixedByteArray(encoding.GetBytes(Inf_Table.username)));
                // パスワード名 256byte (0x4204 , 256)
                Info.AddRange(GetFixedByteArray(encoding.GetBytes(Inf_Table.password)));

                // あまりのスペースを埋める
                Info.AddRange(new byte[3324]);

                // 統計Init関数コール回数(0x5000 ,4)
                Info.AddRange(BitConverter.GetBytes(Inf_Table.init_cnt));
                // 統計Open関数コール回数(0x5004 ,4)
                Info.AddRange(BitConverter.GetBytes(Inf_Table.open_cnt));
                // 統計Close関数コール回数(0x5008 ,4)
                Info.AddRange(BitConverter.GetBytes(Inf_Table.close_cnt));
                // 統計Read関数コール回数(0x500C ,4)
                Info.AddRange(BitConverter.GetBytes(Inf_Table.read_cnt));
                // 統計Write関数コール回数(0x5010 ,4)
                Info.AddRange(BitConverter.GetBytes(Inf_Table.write_cnt));
                
                //--------------------------------------------------
                // 管理データの指定部分ピックアップ
                //--------------------------------------------------
                Info.Skip(offs).Take(size).ToArray().CopyTo(buffer, 0);
            }
            catch { }

            return 0;
        }

    }
}
