﻿////
/// Copyright(C) 2020 Micronet Co. All Rights Reserved.
//
/// <summary>
/// 通信系構造の汎用フレームワーククラス
/// ・送信キュー積込み/取り出し TxPush()/TxPop()
/// ・受信キュー積込み/取り出し RxPush()/RxPop()
/// ・接続完了イベント発行/待機 ConnectedWait()
/// ・受信イベント発行/待機（切断待機兼用） ReceiveWait()
/// ・切断イベント発行
/// ・送信完了イベント発行/待機 TransmitWait()
/// ・接続試行中フラグ Connecting
/// ・接続中フラグ Connected
/// </summary>
using System;
using System.Collections.Generic;
using System.Net;
using System.Text.RegularExpressions;
using System.Threading;

namespace Rsi
{
    internal class Communication
    {
        // event
        private AutoResetEvent EventConnect;    // サーバに接続完了した際のイベント
        private AutoResetEvent EventReceive;    // サーバ受信、または切断された際のイベント
        private AutoResetEvent EventTransmit;   // 送信に成功した際のイベント
        // 送受信データ用のキューバッファ
        public FifoStream Rx;
        public FifoStream Tx;
        public int RxQueSize = 8092;        // デフォルトサイズ
        public int TxQueSize = 8092;        // デフォルトサイズ
        // 状態フラグ
        public bool Connected;              // 接続確立中は true になります
        public bool Connecting;             // 接続試行中は true になります

        /// <summary>
        /// コンストラクタ初期化
        /// </summary>
        public Communication(int TargetTxQueSize = -1, int TargetRxQueSize = -1)
        {
            Connected = false;
            Connecting = false;
            // 内部イベント生成
            EventConnect = new AutoResetEvent(false);
            EventReceive = new AutoResetEvent(false);
            EventTransmit = new AutoResetEvent(false);
            // 送受信キューの固定サイズ決定(-1ならデフォルト)
            RxQueSize = (TargetRxQueSize > 0) ? TargetRxQueSize : RxQueSize;
            TxQueSize = (TargetTxQueSize > 0) ? TargetTxQueSize : TxQueSize;
            // 送受信キューの確保
            Rx = new FifoStream();
            Tx = new FifoStream();
        }

        /// <summary>
        /// 通信管理のリセット
        /// </summary>
        public void Reset()
        {
            Connected = false;
            Connecting = false;
            EventConnect.Reset();
            EventReceive.Reset();
            EventTransmit.Reset();
            Tx.Clear();
            Rx.Clear();
        }

        /// <summary>
        /// 送信キューに積み込む
        /// </summary>
        /// <param name="buf"></param>
        /// <returns>0=成功、-1=キューがいっぱいです</returns>
        public bool TxPush(byte[] buf)
        {
            // キューの残量が充分なら積み込みます
            int RemainSize = TxQueSize - (int)Tx.Count;
            if (RemainSize >= buf.Length)
            {
                // データを積む
                Tx.Push(buf);
                return true;
            }
            else
                return false;
        }

        /// <summary>
        /// 受信キューに積み込む
        /// </summary>
        /// <param name="buf"></param>
        /// <returns>true=成功、false=キューがいっぱいです</returns>
        public bool RxPush(byte[] buf)
        {
            // キューの残量が充分なら積み込みます
            int RemainSize = RxQueSize - (int)Rx.Count;
            if (RemainSize >= buf.Length)
            {
                // データを積んで受信イベントの発行
                Rx.Push(buf);
                ReceivedEvent();
                return true;
            }
            else
                return false;
        }

        /// <summary>
        /// 送信キューから全部取り出す
        /// </summary>
        /// <param name="size">要求サイズ、0は全部取り出し</param>
        /// <param name="maxsize">最大サイズ、全部取り出しでもこのサイズを超えることはできない</param>
        /// <returns>取り出したサイズ</returns>
        public byte[] TxPop(int size, int maxsize)
        {
            // 送信キューにたまっているサイズ
            int streamlen = Tx.Count;
            // 溜まっていないならnull
            if (streamlen == 0)
                return null;
            // 取り出すサイズを決定(要求サイズ0以下の指定は全取り出し)
            streamlen = (size == 0) ? streamlen : Math.Min(streamlen, size);
            // 但し最大サイズを超えることはできない
            streamlen = (streamlen > maxsize) ? maxsize : streamlen;
            // バッファを新たに確保する
            byte[] buf = new byte[streamlen];
            // 先頭から取り出します
            buf = Tx.Pop(streamlen);
            return buf;
        }

        /// <summary>
        /// 受信キューから取り出す
        /// </summary>
        /// <param name="size">要求サイズ、0は全部取り出し</param>
        /// <param name="maxsize">最大サイズ、全部取り出しでもこのサイズを超えることはできない</param>
        /// <returns>取り出したサイズ</returns>
        public byte[] RxPop(int size ,int maxsize)
        {
            // 受信キューにたまっているサイズ
            int streamlen = Rx.Count;
            // 溜まっていないならnull
            if (streamlen == 0)
                return null;
            // 取り出すサイズを決定(要求サイズ0以下の指定は全取り出し)
            streamlen = (size == 0) ? streamlen : Math.Min(streamlen, size);
            // 但し最大サイズを超えることはできない
            streamlen = (streamlen > maxsize) ? maxsize : streamlen;
            // バッファを新たに確保する
            byte[] buf = new byte[streamlen];
            // 先頭から取り出します
            buf = Rx.Pop(streamlen);
            return buf;
        }

        /// <summary>
        /// 送信完了の待ち
        /// </summary>
        /// <param name="timeout">msec</param>
        /// <returns></returns>
        public int TxWait(int timeout = -1)
        {
            bool result = EventTransmit.WaitOne(timeout);
            // 切断はブレイク
            if (Connected == false)
                return -2;
            // タイムアウトはブレイク
            if (result == false)
                return -1;
            // 正常
            return 0;
        }

        /// <summary>
        /// 受信完了または切断の待ち
        /// </summary>
        /// <param name="timeout">msec</param>
        /// <param name="size">待機したいbyte数、デフォルトは1</param>
        /// <returns>0=正常、-1=タイムアウト、-2=切断</returns>
        public int RxWait(int timeout = -1 ,int size = 0)
        {
            bool result;

            while( Rx.Count < size )
            {
                // 受信/切断イベント待機
                result = EventReceive.WaitOne(timeout);
                // タイムアウトはブレイク
                if (result == false)
                    return -1;
                // 切断はブレイク
                if (Connected == false)
                    return -2;
            }
            // 正常
            return 0;
        }

        /// <summary>
        /// 接続完了の待ち
        /// </summary>
        /// <param name="timeout"></param>
        /// <returns>0=正常、-1=タイムアウト</returns>
        public int ConnectedWait(int timeout)
        {
            bool result = EventConnect.WaitOne(timeout);
            if (result == true)
                return 0;       // normal
            else
                return -1;      // timeout
        }

        /// <summary>
        /// 接続処理開始イベントの発行(実際にはイベント使ってない)
        /// </summary>
        public void ConnectingEvent()
        {
            Reset();
        }

        /// <summary>
        /// 接続されたイベントの発行
        /// </summary>
        public void ConnectedEvent()
        {
            Connected = true;
            Connecting = false;
            EventConnect.Set();
        }

        /// <summary>
        /// 切断されたイベントの発行
        /// </summary>
        public void DisconnectedEvent()
        {
            Connected = false;
            Connecting = false;
            EventReceive.Set();
            EventTransmit.Set();
        }

        /// <summary>
        /// 受信したイベントの発行
        /// </summary>
        public void ReceivedEvent()
        {
            EventReceive.Set();
        }

        /// <summary>
        /// 送信完了したイベントの発行
        /// </summary>
        public void TransmitEvent()
        {
            EventTransmit.Set();
        }

        /// <summary>
        /// 文字列からIPアドレス、ポートの解釈
        /// </summary>
        /// <param name="AddressColonPort">文字列</param>
        /// <returns>IPEndPoint</returns>
        public IPEndPoint TreatIPAddress(string AddressColonPort)
        {
            // TCP接続の使用、IPアドレスの解釈
            IPAddress TcpIp;
            int TcpPort;
            try
            {
                // IPアドレスの解釈を試みる
                Match m = Regex.Match(Convert.ToString(AddressColonPort), @"^(?<IP>[0-9a-zA-Z.]*)(:|$)");
                TcpIp = IPAddress.Parse(m.Groups["IP"].Value);
            }
            catch
            {
                try
                {
                    // IPで解釈できない場合はドメイン名で解釈）
                    Match m = Regex.Match(Convert.ToString(AddressColonPort), @"^(?<IP>*)(:|$)");
                    TcpIp = Dns.GetHostAddresses(m.Groups["IP"].Value)[0];
                }
                catch
                {
                    // それでもだめならデフォルト(ANY)
                    TcpIp = IPAddress.Any;
                }
            }
            try
            {
                // 接続ポート番号の解釈（ポート番号指定があるなら）
                Match m = Regex.Match(Convert.ToString(AddressColonPort), @":(?<PORT>\d+)$");
                TcpPort = int.Parse(m.Groups["PORT"].Value);
            }
            catch
            {
                // ポート番号指定がないならデフォルトになります(ANY)
                TcpPort = 0;
            }
            return new IPEndPoint(TcpIp, TcpPort);
        }
    }

    /// <summary>
    /// 高速スレッドセーフFIFO
    /// 
    /// Programmed by Kazumi.M
    /// 
    /// Count ... 溜まっているバイト数
    /// void Push(byte[]) ... 積む
    /// byte[] Pop(size) ... 取り出す
    /// void Clear() ... 空にします
    /// </summary>
    internal class FifoStream
    {
        private List<Byte> FifoData = new List<byte>();

        public UInt64 TotalPushBytes;
        public UInt64 TotalPopBytes;
        public int LastPushBytes;
        public int LastPopBytes;

        public FifoStream()
        {
            Clear();
        }

        /// <summary>
        /// カウント数
        /// </summary>
        public int Count
        {
            get
            {
                lock (FifoData)
                {
                    return FifoData.Count;
                }
            }
        }

        /// <summary>
        /// クリア
        /// </summary>
        public void Clear()
        {
            lock (FifoData)
            {
                FifoData.Clear();
                TotalPushBytes = 0;
                TotalPopBytes = 0;
                LastPushBytes = 0;
                LastPopBytes = 0;
            }
        }

        /// <summary>
        /// ブロックPush
        /// </summary>
        /// <param name="u8_Data"></param>
        public void Push(byte[] u8_Data)
        {
            lock (FifoData)
            {
                // 積込み
                FifoData.AddRange(u8_Data);
                // 統計
                LastPushBytes = u8_Data.Length;
                TotalPushBytes += (ulong)LastPushBytes;
            }
        }

        /// <summary>
        /// ブロックPop
        /// </summary>
        /// <param name="s32_Count"></param>
        /// <returns></returns>
        public Byte[] Pop(int s32_Count)
        {
            lock (FifoData)
            {
                // 取り出せない
                if (FifoData.Count < s32_Count)
                    return null;
                // 取り出し
                Byte[] u8_PopData = new byte[s32_Count];
                FifoData.CopyTo(0, u8_PopData, 0, s32_Count);
                FifoData.RemoveRange(0, s32_Count);
                // 統計
                LastPopBytes = s32_Count;
                TotalPopBytes += (ulong)LastPopBytes;
                return u8_PopData;
            }
        }
    }
}
