﻿////
/// Copyright(C) 2020 Micronet Co. All Rights Reserved.
//
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Threading.Tasks;

namespace Rsi
{
    public partial class HiTCPC : Rsi
    {
        //--------------------------------------------------
        // クラスメンバ変数の定義
        //--------------------------------------------------
        private InfomationArea Inf_Table;       // 情報領域テーブル

        // 通信監理フレームワーク
        private Communication comm;

        // TCP Client protocol
        private Socket TcpClient;                       // ソケットIF
        private IPEndPoint ServerAddress;               // サーバアドレス(IP+ポート)
        private IPEndPoint LocalAddress;                // ローカルホストアドレス(IP+ポート)

        // オプション設定
        private bool KeepAliveEnable;
        private int KeepAliveTime;
        private int KeepAliveInterval;
        private bool ReUserAddrEnable;

        // TCP非同期メッセージ
        private SocketAsyncEventArgs TxAsyncEventArgs;  // 登り方向のソケット通信イベント
        private SocketAsyncEventArgs RxAsyncEventArgs;  // 下り方向のソケット通信イベント
        byte[] Rx = new byte[2000];                     // イベントが扱うバッファ領域

        // Statistics
        public int ConnectCounter = 0;      // 接続成功回数
        public int DisconnectCounter = 0;   // 切断発生回数

        //--------------------------------------------------
        // ステータスコード宣言
        //--------------------------------------------------

        /* 区分:その他のエラー */


        /// <summary>
        /// コンストラクタ
        /// </summary>
        public HiTCPC() : base()
        {
            // 情報領域テーブルの生成・初期化

            // デフォルト設定値

            // 通信監理フレームワークの初期化
            comm = new Communication();
        }

        /// <summary>
        /// TCPイベントハンドラ
        /// </summary>
        /// <returns></returns>
        internal void fnTCPCEvent(object sender, SocketAsyncEventArgs e)
        {
            switch (e.LastOperation)
            {
                // 接続のイベントが発生
                case SocketAsyncOperation.Connect:
                    if (e.SocketError == SocketError.Success)
                    {
                        Trace.WriteLine("### CONNECTED WITH SERVER ###");
                        // 次の受信イベントを出しておく
                        RxAsyncEventArgs.SetBuffer(Rx, 0, Rx.Length);
                        Task.Run(() => TcpClient.ReceiveAsync(RxAsyncEventArgs));
                        // 通信状態の制御
                        ConnectCounter++;
                        comm.ConnectedEvent();
                    }
                    break;

                // 送信完了のイベントが発生
                case SocketAsyncOperation.Send:
                    if (e.SocketError == SocketError.Success)
                    {
                        comm.TransmitEvent();
                    }
                    break;

                // 受信完了のイベントが発生
                case SocketAsyncOperation.Receive:
                    // TCPメッセージの取り出し
                    var RxSize = e.BytesTransferred;
                    // 切断の判断
                    if (RxSize == 0)
                        goto disconnect;
                    // 受信した分をキューに書出し
                    Trace.WriteLine("### RECEIVED APPLICATION MESSAGE ###");
                    comm.RxPush(e.Buffer.Take(RxSize).ToArray());
                    // 次の受信イベントを出しておく
                    RxAsyncEventArgs.SetBuffer(Rx, 0, Rx.Length);
                    Task.Run(() => TcpClient.ReceiveAsync(RxAsyncEventArgs));
                    break;

                // 切断のイベントが発生
                case SocketAsyncOperation.Disconnect:
                disconnect:
                    Trace.WriteLine("### DISCONNECTED FROM SERVER ###");
                    // 切断発生回数を増やす
                    if (ConnectCounter > 0)
                        DisconnectCounter++;
                    // 通信状態の制御
                    comm.DisconnectedEvent();
                    //
                    EventStop();
                    break;
            }
        }

        /// <summary>
        /// ターゲットパラメータの解釈処理
        /// </summary>
        /// <param name="target">ターゲットパラメータ文字列</param>
        /// <returns></returns>
        internal UInt32 fnTreatTarget(string target)
        {
            UInt16 count = 0;
            foreach (string s in ItemValues.StringSplitOptions(target, '|')) // '|' で区切られたtarget文字列を処理する
            {
                // 一番先頭の文節は接続先として認識されます
                if (count++ == 0)
                {
                    // IPアドレス：ポート番号を解釈します
                    ServerAddress = comm.TreatIPAddress(s);
                }
            }
            return ER_HI_OK;
        }

        /// <summary>
        /// コンフィグパラメータの解釈処理
        /// </summary>
        /// <param name="config">コンフィグパラメータ文字列</param>
        /// <returns></returns>
        internal UInt32 fnTreatConfig(string config)
        {
            foreach (string s in ItemValues.StringSplitOptions(config, '|')) // '|' で区切られたconfig文字列を処理する
            {
                // "item=value"の書式となるオプション設定値があれば処理していきます
                ItemValue setting = new ItemValue(s);
                if (setting.item.Equals("KeepAlive"))
                {
                    // KeepAlive有効
                    KeepAliveEnable = true;
                    continue;
                }
                if (setting.item.Equals("KeepAliveTime"))
                {
                    KeepAliveEnable = true;
                    KeepAliveTime = Convert.ToInt32(setting.value);
                    continue;
                }
                if (setting.item.Equals("KeepAliveInterval"))
                {
                    KeepAliveEnable = true;
                    KeepAliveInterval = Convert.ToInt32(setting.value);
                    continue;
                }
                if (setting.item.Equals("LocalIP"))
                {
                    // IPアドレス：ポート番号を解釈します
                    LocalAddress = comm.TreatIPAddress(setting.value);
                    continue;
                }
                if( setting.item.Equals("ReUseAddr"))
                {
                    // ReuseAddr有効
                    ReUserAddrEnable = true;
                }
            }
            return ER_HI_OK;
        }

        /// <summary>
        /// 初期化の要求
        /// </summary>
        /// <returns></returns>
        internal bool fnInitRequest()
        {
            return true;
        }

        /// <summary>
        /// 初期化の完了待ち
        /// </summary>
        /// <param name="dwTimeout">msec</param>
        /// <returns></returns>
        internal int fnInitWait(UInt32 dwTimeout)
        {
            return 0;
        }

        /// <summary>
        /// 接続のリクエスト
        /// </summary>
        /// <returns></returns>
        internal bool fnOpenRequest()
        {
            // TCPC コンポーネントのインスタンスを生成します
            TcpClient?.Dispose();
            TcpClient = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            // イベントの再生成
            EventStart();
            // KeepAlive 設定
            if (KeepAliveEnable == true)
            {
                TcpClient.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.KeepAlive, true);
                // 構造体 tcp_keepalive.
                List<byte> tcp_keepalive = new List<byte>();
                tcp_keepalive.AddRange(BitConverter.GetBytes((Int32)1));    //onoffスイッチ.
                tcp_keepalive.AddRange(BitConverter.GetBytes((Int32)KeepAliveTime));
                tcp_keepalive.AddRange(BitConverter.GetBytes((Int32)KeepAliveInterval));
                TcpClient.IOControl(IOControlCode.KeepAliveValues, tcp_keepalive.ToArray(), null);
            }
            // Local設定
            if (LocalAddress != null)
            {
                TcpClient.Bind(LocalAddress);
            }
            // Reuse設定
            if (ReUserAddrEnable == true)
            {
                TcpClient.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, true);
            }
            // イベントの準備と接続の要求
            TxAsyncEventArgs.SetBuffer(0, 0);
            TxAsyncEventArgs.RemoteEndPoint = ServerAddress;
            Task.Run(() => TcpClient.ConnectAsync(TxAsyncEventArgs));
            return true;
        }

        /// <summary>
        /// 接続完了するまで待ち
        /// </summary>
        /// <param name="dwTimeout">msec、WAIT_FOREVERは無限待ち</param>
        /// <returns></returns>
        internal int fnOpenWait(UInt32 dwTimeout)
        {
            int inttimeout = (dwTimeout > (UInt32)Int32.MaxValue) ? -1 : Convert.ToInt32(dwTimeout);
            return comm.ConnectedWait(inttimeout);
        }

        /// <summary>
        /// 切断のリクエスト
        /// </summary>
        /// <returns></returns>
        internal bool fnCloseRequest()
        {
            // 現在のステータス
            comm.DisconnectedEvent();
            // 切断
            if (TcpClient != null)
            {
                TxAsyncEventArgs.SetBuffer(0, 0);
                Task.Run(() => TcpClient.DisconnectAsync(TxAsyncEventArgs));
                //// イベントの停止
                //EventStop();
                //
                TcpClient?.Dispose();
                TcpClient = null;
            }
            return true;
        }

        /// <summary>
        /// 切断完了の待ち
        /// </summary>
        /// <param name="dwTimeout">msec</param>
        /// <returns></returns>
        internal int fnCloseWait(UInt32 dwTimeout)
        {
            return 0;
        }

        /// <summary>
        /// 読み込みのリクエスト
        /// </summary>
        /// <param name="message"></param>
        /// <param name="isize"></param>
        /// <returns></returns>
        internal int fnReadRequest( ref byte[] message, int isize)
        {
            // 切断されている場合
            if (comm.Connected == false)
                return -2;

            // 受信キューから取り出します
            byte[] input = comm.RxPop(isize ,message.Length-2);
            byte[] len = BitConverter.GetBytes( Convert.ToUInt16(input.Length));
            len.CopyTo(message, 0);
            input.CopyTo(message, 2);
            return input.Length;
        }

        /// <summary>
        /// 受信又は指定サイズの受信、あるいは切断されるまで待ち
        /// </summary>
        /// <param name="dwTimeout">msec</param>
        /// <returns></returns>
        internal int fnReadWait(UInt32 dwTimeout, int size)
        {
            int inttimeout = (dwTimeout > (UInt32)Int32.MaxValue) ? -1 : Convert.ToInt32(dwTimeout);
            return comm.RxWait(inttimeout, (size == 0) ? 1 : size);
        }

        /// <summary>
        /// 書き込みのリクエスト
        /// </summary>
        /// <param name="message"></param>
        /// <param name="isize"></param>
        /// <returns></returns>
        internal int fnWriteRequest( byte[] message, int isize)
        {
            // 切断されている場合
            if (comm.Connected == false)
                return -2;

            // 送信キュー（実際には使っておらずカウントしているのみ）
            comm.TxPush(message);
            comm.TxPop(0, message.Length);
            // イベントの準備と送信要求
            TxAsyncEventArgs.SetBuffer(message, 0, isize);
            Task.Run(() => TcpClient.SendAsync(TxAsyncEventArgs));
            return 0;
        }

        /// <summary>
        /// 書き込み完了の待ち
        /// </summary>
        /// <param name="dwTimeout">msec</param>
        /// <returns></returns>
        internal int fnWriteWait(UInt32 dwTimeout)
        {
            int inttimeout = (dwTimeout > (UInt32)Int32.MaxValue) ? -1 : Convert.ToInt32(dwTimeout);
            return comm.TxWait(inttimeout);
        }

        internal int fnReadInfo( ref byte[] buffer ,int offs ,int size )
        {
            try
            {
                //--------------------------------------------------
                // 管理データの組み立て
                //--------------------------------------------------
                List<byte> Info = new List<byte>();
                // 接続ステータス(0x4000 ,4)
                var status = (comm.Connected) ? 2 : (comm.Connecting) ? 1 : 0;
                Info.AddRange(BitConverter.GetBytes(status));
                // 現在受信可能なバイト数(0x4004 ,4)
                Info.AddRange(BitConverter.GetBytes(comm.Rx.Count));
                // トータル送信バイト数(0x4008 ,8)
                Info.AddRange(BitConverter.GetBytes(comm.Tx.TotalPushBytes));
                // トータル受信バイト数(0x4010 ,8)
                Info.AddRange(BitConverter.GetBytes(comm.Rx.TotalPushBytes));
                // 最終の送信バイト数(0x4018 ,4)
                Info.AddRange(BitConverter.GetBytes(comm.Tx.LastPushBytes));
                // 最終の受信バイト数(0x401C ,4)
                Info.AddRange(BitConverter.GetBytes(comm.Rx.LastPushBytes));
                // 統計Init関数コール回数(0x4020 ,4)
                Info.AddRange(BitConverter.GetBytes(Inf_Table.init_cnt));
                // 統計Open関数コール回数(0x4024 ,4)
                Info.AddRange(BitConverter.GetBytes(Inf_Table.open_cnt));
                // 統計Close関数コール回数(0x4028 ,4)
                Info.AddRange(BitConverter.GetBytes(Inf_Table.close_cnt));
                // 統計Read関数コール回数(0x402C ,4)
                Info.AddRange(BitConverter.GetBytes(Inf_Table.read_cnt));
                // 統計Write関数コール回数(0x4030 ,4)
                Info.AddRange(BitConverter.GetBytes(Inf_Table.write_cnt));

                //--------------------------------------------------
                // 管理データの指定部分ピックアップ
                //--------------------------------------------------
                Info.Skip(offs).Take(size).ToArray().CopyTo(buffer, 0);
            }
            catch { }

            return 0;
        }

        /// <summary>
        /// イベント通知の初期化
        /// </summary>
        internal void EventStart()
        {
            TxAsyncEventArgs = new SocketAsyncEventArgs();
            RxAsyncEventArgs = new SocketAsyncEventArgs();
            TxAsyncEventArgs.Completed += new EventHandler<SocketAsyncEventArgs>(fnTCPCEvent);
            RxAsyncEventArgs.Completed += new EventHandler<SocketAsyncEventArgs>(fnTCPCEvent);
        }

        /// <summary>
        /// イベント通知の削除
        /// </summary>
        internal void EventStop()
        {
            TxAsyncEventArgs?.Dispose();
            TxAsyncEventArgs = null;
            RxAsyncEventArgs?.Dispose();
            RxAsyncEventArgs = null;
        }

    }
}
